import re
from .base import BaseDevice

VALID_PORTS = re.compile(r"^(ge|xe)-")

ADMIN_STATUS = {
    "up": "enabled",
    "down": "disabled",
}
OPER_STATUS = {"up": "up", "down": "down"}
DUPLEX = {
    "Half-duplex": "half",
    "Full-duplex": "full",
}
SPEED = {
    "Auto": "auto",
    "1000mbps": "1000",
    "100mbps": "100",
    "10mbps": "10",
}


class Junos(BaseDevice):
    def get_interface_details(self, interface=False):
        output = {}

        rpc = (
            self.device.rpc.jrpc__rpc_get_interface_information.get_interface_information
        )
        inp = rpc.get_input()

        self.log.info(f"interface is {interface}")

        # First we're getting interface information
        if interface:
            inp.interface_name = interface
        try:
            reply = rpc(inp)
        except:
            self.log.info(f"request failed")
            return []

        # save the fields we care about in our output as normalized data
        for i in reply.interface_information.physical_interface:

            self.log.info(f"processing port {i.name}")
            if not (VALID_PORTS.match(i.name)):
                continue

            if (not (interface)) or (interface and interface == i.name):
                output[i.name] = {
                    "admin_status": ADMIN_STATUS[i.admin_status],
                    "oper_status": OPER_STATUS[i.oper_status],
                    "speed": SPEED.get(i.speed, "auto"),
                    "duplex": DUPLEX.get(i.link_mode, "auto"),
                }

        return output

    def get_mac_table(self, mac=False):
        """
        Gathers mac address table data from a junos device
        and normalizes the output into a list. Each list entry is a dict
        with the interface name, vlan id, and mac address.
        """

        rpc = (
            self.device.rpc.jrpc__rpc_get_ethernet_switching_table_information.get_ethernet_switching_table_information
        )
        inp = rpc.get_input()
        inp.brief.create()
        if mac:
            mac = self.normalize_mac(mac)
            inp.address = mac
        try:
            reply = rpc(inp)
        except:
            return []

        mac_table = reply.l2ng_l2ald_rtb_macdb.l2ng_l2ald_mac_entry_vlan
        output = []

        if not (mac_table):
            return output
        else:
            mac_table = mac_table[0]

        for i in mac_table.l2ng_mac_entry:

            interface = i.l2ng_l2_mac_logical_interface.replace(".0", "")
            if not (VALID_PORTS.match(interface)):
                continue

            vlan_id = self.device.config.configuration.vlans.vlan[
                i.l2ng_l2_mac_vlan_name
            ].vlan_id
            address = i.l2ng_l2_mac_address

            if (not (mac)) or (mac and mac == address):
                output.append(
                    {
                        "interface": interface,
                        "vlan-id": vlan_id,
                        "address": address,
                    }
                )

        return output

    def get_arp_table(self, ip=False):

        rpc = (
            self.device.rpc.jrpc__rpc_get_arp_table_information.get_arp_table_information
        )
        inp = rpc.get_input()
        if ip:
            inp.hostname = ip

        try:
            reply = rpc(inp)
        except:
            return []

        output = []
        for a in reply.arp_table_information.arp_table_entry:

            if (not (ip)) or (ip and ip == a.ip_address):
                output.append(
                    {
                        "ip": a.ip_address,
                        "mac": a.mac_address,
                    }
                )

        return output


class Junos12(Junos):
    def get_mac_table(self, mac=False):
        """
        Gathers mac address table data from an old junos (12.x) device
        and normalizes the output into a list. Each list entry is a dict
        with the interface name, vlan id, and mac address.
        """

        rpc = (
            self.device.rpc.jrpc__rpc_get_ethernet_switching_table_information.get_ethernet_switching_table_information
        )
        inp = rpc.get_input()
        if mac:
            mac = self.normalize_mac(mac)
            inp.address = mac
        try:
            reply = rpc.get_ethernet_switching_table_information(inp)
        except:
            return []

        mac_table = (
            reply.ethernet_switching_table_information.ethernet_switching_table.mac_table_entry
        )

        output = []
        for mac in mac_table:

            interface = mac.mac_interfaces_list.mac_interfaces.replace(".0", "")
            if (mac.mac_type != "Learn") or not (VALID_PORTS.match(interface)):
                continue

            if not (mac) or (mac == mac.mac_address):
                output.append(
                    {
                        "interface": mac.mac_interfaces_list.mac_interfaces.replace(
                            ".0", ""
                        ),
                        "vlan-id": self.device.config.configuration.vlans.vlan[
                            mac.mac_vlan
                        ].vlan_id,
                        "address": mac.mac_address,
                    }
                )

        return output



import re
import logging
from typing import Any

import ncs


class BaseDevice:
    def __init__(self, device: ncs.maagic.Node, log: logging.Logger):
        self.device = device
        self.log = log

    @staticmethod
    def normalize_mac(mac: str) -> str:
        mac = re.sub(r"[\.\-\:]", "", mac)
        return f"{mac[0:2]}:{mac[2:4]}:{mac[4:6]}:{mac[6:8]}:{mac[8:10]}:{mac[10:12]}"

    def _run_cmd(self, command: str) -> Any:
        raise NotImplementedError

    def get_mac_table(self, mac=False) -> Any:
        raise NotImplementedError

    def get_arp_table(self, ip=False) -> Any:
        raise NotImplementedError

    def get_interface_details(self, interface=False) -> Any:
        raise NotImplementedError

    def get_bfd_neighbors(self, interface=False) -> Any:
        raise NotImplementedError

    def get_ospf_neighbors(self, interface=False) -> Any:
        raise NotImplementedError

    def get_lldp_neighbors(self, interface=False) -> Any:
        raise NotImplementedError

    def get_transceiver_details(self, interface=False) -> Any:
        raise NotImplementedError



import re
from .base import BaseDevice

VALID_PORTS = re.compile(r"^(Fa|Gi|Te)")
VALID_MAC = re.compile(r"[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}")

# note that since we're using 'show int status',
# admin and oper status leverage the same output column, whose values can be
## connected (admin up, oper up)
## notconnect (admin up, oper down)
## disabled (admin down, oper down)
## suspended (admin up, oper down)
## err-disabled (admin up, oper down)
ADMIN_STATUS = {
    "connected": "enabled",
    "notconnect": "enabled",
    "disabled": "disabled",
    "suspended": "enabled",
    "err-disabled": "enabled",
}
OPER_STATUS = {
    "connected": "up",
    "notconnect": "down",
    "disabled": "down",
    "suspended": "down",
    "err-disabled": "down",
}
DUPLEX = {
    "half": "half",
    "full": "full",
    "auto": "full",
    "a-half": "half",
    "a-full": "full",
}
SPEED = {
    "10": "10",
    "100": "100",
    "1000": "1000",
    "auto": "auto",
    "a-10": "10",
    "a-100": "100",
    "a-1000": "1000",
}


class IOS(BaseDevice):
    def get_interface_details(self, interface=False):
        """
        Gathers interface operational data from an ios device by parsing
        'show interfaces status'

        If you provide an interface name to the function it will
        only return data found for that interface.
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        inp.args = (
            [f"interfaces {interface} status"] if interface else ["interfaces status"]
        )
        reply = show.request(inp)

        output = {}
        for line in reply.result.split("\n"):

            i_name = line[0:10].rstrip()
            if not (VALID_PORTS.match(i_name)):
                continue

            if (not (interface)) or (interface and interface == i_name):
                output[i_name] = {
                    "admin_status": ADMIN_STATUS[line[29:42].strip()],
                    "oper_status": OPER_STATUS[line[29:42].strip()],
                    "speed": SPEED[line[60:67].strip()],
                    "duplex": DUPLEX[line[53:60].strip()],
                }

        return output

    def get_arp_table(self, ip=False):
        """
        Gathers arp data from an ios device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        output = []

        # first we'll look in the main table
        inp.args = [f"ip arp {ip}"] if ip else [f"ip arp"]
        inp = show.get_input()
        reply = show.request(inp)
        output.extend(self._parse_arp_reply(reply, ip))

        # Now we'll look in all the vrfs - ios makes you run a separate
        # lookup command for every vrf (so rude)
        vrfs = [vrf.name for vrf in self.device.config.vrf.definition]
        for vrf in vrfs:
            inp.args = [f"ip arp vrf {vrf} {ip}"] if ip else [f"ip arp vrf {vrf}"]
            reply = show.request(inp)
            output.extend(self._parse_arp_reply(reply, ip))

        return output

    def get_mac_table(self, mac=False):
        """
        Parses mac address table data from an ios device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        if mac:
            inp.args = [f"mac address-table address {mac}"]
        else:
            inp.args = ["mac address-table dynamic"]
        reply = show.request(inp)

        self.log.info(f"results {reply.result}")
        output = []
        for line in reply.result.split("\n"):

            interface = line[38:46].strip()
            self.log.info(f"interface {interface}")
            if not (VALID_PORTS.match(interface)):
                continue

            vlan_id = line[0:8].strip()
            address = line[8:26].strip()

            # TODO: fix mac comparison
            if (not (mac)) or (mac and mac == self.normalize_mac(address)):
                output.append(
                    {
                        "interface": interface,
                        "vlan-id": vlan_id,
                        "address": self.normalize_mac(address),
                    }
                )

        return output

    def _parse_arp_reply(self, reply, ip):

        for line in reply.result.split("\n"):

            arp_ip = line[10:35].strip()
            arp_mac = line[38:52].strip()

            if not (VALID_MAC.match(arp_mac)):
                continue

            if (not (ip)) or (ip and arp_ip == ip):
                output.append(
                    {
                        "mac": arp_mac,
                        "ip": arp_ip,
                    }
                )

        return output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['umnet_pyncs', 'umnet_pyncs.state', 'umnet_pyncs.state.models']

package_data = \
{'': ['*']}

install_requires = \
['multiprocessing-logging>=0.3.3,<0.4.0', 'ntc-templates>=3.0.0,<4.0.0']

setup_kwargs = {
    'name': 'umnet-pyncs',
    'version': '0.1.5',
    'description': 'custom python module for NCS helpers',
    'long_description': '## Overview\n\nThis module is intended to be installed on the production NCS nodes and imported in other services/actions that need to gather state from the network.  It uses the NCS device manager and the standard python `multiprocessing` library to connect to devices in-parallel and issue commands, returning results as structured data.\n\n## Usage information\n\nBasic usage example in an NCS callback:\n\n``` python\nfrom umnet_ncs.state import StateManager\n...\n\n\nclass DemoAction(Action):\n    @Action.action\n    def cb_action(self, uinfo, name, kp, input, output, trans):\n        ...\n        with StateManager() as m:\n            interfaces = m.get_state(al_devices, ["get-interface-details"])\n            arp = m.get_state(dl_devices, ["get-arp-table"])\n            ...\n```\n\n## Supported commands\n\nWe attempt to normalize the output of each command based on how it is implemented.  For example, we might default to just directly returning the data as-parsed by the `ntc_templates` module, or try and emulate for e.g. junos devices where we might have direct access to NETCONF RPCs via the NCS device manager.\n\nCurrently supported commands are:\n- `get-mac-table`\n- `get-arp-table`\n- `get-interface-details`\n- `get-transciever-details`\n\n## TODO\nThe models are mostly taken verbatim from the `netsplash` NCS package, and additional methods were added on top to support the [stats.py](https://gitlab.umich.edu/its-inf-net/umnet-ncs-dev/-/blob/master/packages/umnet-backbone/python/fabric/stats.py) module.\n',
    'author': 'Nick Grundler',
    'author_email': 'grundler@umich.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re
from typing import List, Dict, Any

from ntc_templates.parse import parse_output

from .base import BaseDevice

VALID_PORTS = re.compile(r"^Ethernet")
VALID_MAC = re.compile(r"[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}")

ADMIN_STATUS = {
    "connected": "enabled",
    "suspndByV": "enabled",
    "notconnec": "enabled",
    "disabled": "disabled",
    "sfpAbsent": "enabled",
    "xcvrAbsen": "enabled",
    "suspnd": "enabled",
    "suspended": "enabled",
}
OPER_STATUS = {
    "connected": "up",
    "suspndByV": "down",
    "notconnec": "down",
    "disabled": "down",
    "sfpAbsent": "down",
    "xcvrAbsen": "down",
    "suspnd": "down",
    "suspended": "down",
}
DUPLEX = {
    "half": "half",
    "full": "full",
    "auto": "full",
}
SPEED = {
    "10": "10",
    "100": "100",
    "1000": "1000",
    "auto": "auto",
    "a-10": "10",
    "a-100": "100",
    "a-1000": "1000",
    "a-10G": "10000",
}


class NXOS(BaseDevice):
    def _run_cmd(self, command: str) -> str:
        """
        use NCS live-status exec to issue a raw show command towards a device.
        platform-specific models are expected to parse this output and return
        structured data.

        :param command: string CLI command, e.g. 'show interface status'
        :returns: raw string output from the device
        """
        show = self.device.live_status.__getitem__("exec").show
        inp = show.get_input()
        inp.args = [command]

        return show.request(inp)

    def get_interface_details(self, interface=False) -> List[Dict]:
        """
        Gathers interface operational data from an NXOS device by parsing
        'show interfaces status'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface_status
        """
        # ncs automatically adds the 'show' to the front of the cmd
        self.log.debug(f"sending 'show interface status' to {self.device.name}")
        command = f"interface {interface} status" if interface else "interface status"
        reply = self._run_cmd(command)
        self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_bfd_neighbors(self) -> List:
        """
        Gathers bfd operational data from an NXOS device by parsing 'show bfd
        neighbors'

        see the ntc_templates test data for this template for details on output structure:
        https://gitlab.umich.edu/grundler/ntc-templates/-/tree/add-nxos-show-bfd-neighbors/tests/cisco_nxos/show_bfd_neighbors
        """
        self.log.debug(f"sending 'show bfd neighbors' to {self.device.name}")
        command = "bfd neighbors"
        reply = self._run_cmd(command)
        self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_ospf_neighbors(self) -> List:
        """
        Gathers live ospf operational data from an NXOS device by parsing 'show
        ip ospf neighbor'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_ip_ospf_neighbor
        """
        self.log.debug(f"sending 'show ip ospf neighbor' to {self.device.name}")
        command = "ip ospf neighbor"
        reply = self._run_cmd(command)
        self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_lldp_neighbors(self) -> List:
        """
        Gathers live operational data about physically connected devices from an
        NXOS device by parsing 'show lldp neighbors'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_lldp_neighbors
        """
        self.log.debug(f"sending 'show ip ospf neighbors' to {self.device.name}")
        command = "lldp neighbors"
        reply = self._run_cmd(command)
        self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_transceiver_details(self) -> List:
        """
        Gathers live operational state DOM information from an NXOS device by
        parsing 'show interface tranceiver details'

        see the ntc_templates test data for this template for details on output structure:
        https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface_transceiver_details
        """
        self.log.debug(f"sending 'show ip ospf neighbors' to {self.device.name}")
        command = "interface transceiver details"
        reply = self._run_cmd(command)
        self.log.debug(f" <<<{ self.device.name }>>>: {reply.result}")

        return parse_output(
            platform="cisco_nxos",
            command="show " + command,
            data=reply.result,
        )

    def get_arp_table(self, ip=False):
        raise NotImplementedError

    def get_mac_table(self, mac=False):
        raise NotImplementedError



from ..models.junos import Junos, Junos12
from ..models.nxos import NXOS
from ..models.ios import IOS

from ncs.application import get_ned_id


def get_parser(device):
    """
    instantiate and return the appropriate device state parsing model.
    """
    ned_id = get_ned_id(device)
    if ned_id.startswith("cisco-ios"):
        return IOSParser(root, device, log)
    elif ned_id.startswith("juniper-junos12"):
        return Junos12Parser(root, device, log)
    elif ned_id.startswith("juniper-junos17"):
        return Junos17Parser(root, device, log)
    else:
        raise NotImplementedError(f"Unsupported ned-id {ned_id}")
